@@ -1,7 +1,7 @@
 # Base class for experiments
 #
 # Copyright (C) 2016--2021 Simon Dobson
-# 
+#
 # This file is part of epyc, experiment management in Python.
 #
 # epyc is free software: you can redistribute it and/or modify
@@ -17,14 +17,14 @@
 # You should have received a copy of the GNU General Public License
 # along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.
 
+import sys
 from datetime import datetime
 import traceback
-import sys
+from typing import Set, Dict, Union, List, Any, Final
 if sys.version_info >= (3, 8):
-    from typing import Set, Dict, Union, List, Any, Final
+    from typing import Final
 else:
-    # backwards compatibility with Python 35, Python36, and Python37 
-    from typing import Set, Dict, Union, List, Any
+    # backwards compatibility with Python35, Python36, and Python37
     from typing_extensions import Final
 
 # The type of results dicts
@@ -51,25 +51,25 @@ class Experiment(object):
     """
 
     # Top-level structure for results dicts
-    METADATA : Final[str] = 'metadata'                 #: Results dict key for metadata values, mainly timing.
-    PARAMETERS : Final[str] = 'parameters'             #: Results dict key for describing the point in the parameter space the experiment ran on.
-    RESULTS : Final[str] = 'results'                   #: Results dict key for the experimental results generated at the experiment's parameter point.
+    METADATA: Final[str] = 'metadata'                 #: Results dict key for metadata values, mainly timing.
+    PARAMETERS: Final[str] = 'parameters'             #: Results dict key for describing the point in the parameter space the experiment ran on.
+    RESULTS: Final[str] = 'results'                   #: Results dict key for the experimental results generated at the experiment's parameter point.
 
     # Standard metadata elements reported
-    EXPERIMENT : Final[str] = 'epyc.experiment.classname'              #: Metadata element for storing the class name of the experiment.
-    START_TIME : Final[str] = 'epyc.experiment.start_time'             #: Metadata element for the datetime experiment started.
-    END_TIME : Final[str] = 'epyc.experiment.end_time'                 #: Metadata element for the datetime experiment ended.
-    ELAPSED_TIME : Final[str] = 'epyc.experiment.elapsed_time'         #: Metadata element for the time experiment took overall in seconds.
-    SETUP_TIME : Final[str] = 'epyc.experiment.setup_time'             #: Metadata element for the time spent on setup in seconds.
-    EXPERIMENT_TIME : Final[str] = 'epyc.experiment.experiment_time'   #: Metadata element for the time spent on experiment itself in seconds.
-    TEARDOWN_TIME : Final[str] = 'epyc.experiment.teardown_time'       #: Metadata element for the time spent on teardown in seconds.
-    STATUS : Final[str] = 'epyc.experiment.status'                     #: Metadata element that will be True if experiment completed successfully, False otherwise.
-    EXCEPTION : Final[str] = 'epyc.experiment.exception'               #: Metadata element containing the exception thrown if experiment failed.
-    TRACEBACK : Final[str] = 'epyc.experiment.traceback'               #: Metadata element containing the traceback from the exception (as a string).
+    EXPERIMENT: Final[str] = 'epyc.experiment.classname'              #: Metadata element for storing the class name of the experiment.
+    START_TIME: Final[str] = 'epyc.experiment.start_time'             #: Metadata element for the datetime experiment started.
+    END_TIME: Final[str] = 'epyc.experiment.end_time'                 #: Metadata element for the datetime experiment ended.
+    ELAPSED_TIME: Final[str] = 'epyc.experiment.elapsed_time'         #: Metadata element for the time experiment took overall in seconds.
+    SETUP_TIME: Final[str] = 'epyc.experiment.setup_time'             #: Metadata element for the time spent on setup in seconds.
+    EXPERIMENT_TIME: Final[str] = 'epyc.experiment.experiment_time'   #: Metadata element for the time spent on experiment itself in seconds.
+    TEARDOWN_TIME: Final[str] = 'epyc.experiment.teardown_time'       #: Metadata element for the time spent on teardown in seconds.
+    STATUS: Final[str] = 'epyc.experiment.status'                     #: Metadata element that will be True if experiment completed successfully, False otherwise.
+    EXCEPTION: Final[str] = 'epyc.experiment.exception'               #: Metadata element containing the exception thrown if experiment failed.
+    TRACEBACK: Final[str] = 'epyc.experiment.traceback'               #: Metadata element containing the traceback from the exception (as a string).
 
     # The above, collected together
-    StandardMetadata : Set[str]                        #: The standard metadata elements to always capture.  
-    StandardMetadataTypes : Dict[str, type]            #: Type mapping for standard metadata.
+    StandardMetadata: Set[str]                        #: The standard metadata elements to always capture.
+    StandardMetadataTypes: Dict[str, type]            #: Type mapping for standard metadata.
 
     def __init__(self):
         self._metadata : Dict[str, Any] = dict()
@@ -89,7 +89,7 @@ class Experiment(object):
                                      cls.STATUS,
                                      cls.EXCEPTION,
                                      cls.TRACEBACK
-                                   ]) 
+                                   ])
         cls.StandardMetadataTypes = { cls.EXPERIMENT: str,
                                       cls.START_TIME: datetime,
                                       cls.END_TIME: datetime,
@@ -129,7 +129,7 @@ class Experiment(object):
         self.configure(params)
         return self
 
-    def configure(self, params : Dict[str, Any]):
+    def configure(self, params: Dict[str, Any]):
         """Configure the experiment for the given parameters.
         The default stores the parameters for later use. Be sure
         to call this base method when overriding.
@@ -146,7 +146,7 @@ class Experiment(object):
 
     # ---------- Running the experiment ----------
 
-    def setUp(self, params : Dict[str, Any]):
+    def setUp(self, params: Dict[str, Any]):
         """Set up the experiment. Default does nothing.
 
         :param params: the parameters of the experiment"""
@@ -156,55 +156,67 @@ class Experiment(object):
         """Tear down the experiment. Default does nothing."""
         pass
 
-    def do(self, params : Dict[str, Any]) -> Union[Dict[str, Any], List[ResultsDict]]:
-        """Do the body of the experiment. This should be overridden
-        by sub-classes. Default does nothing.
+    def do(self, params: Dict[str, Any]) -> Union[Dict[str, Any], List[ResultsDict]]:
+        """Do the body of the experiment. This should be overridden by
+        sub-classes. Default does nothing.
 
         An experiment can return two types of results:
 
-        - a dict mapping names to values for experimental results
-        - a list of :term:`results dict`s, which represent fully-formed experiments
+        - a dict mapping names to values for experimental results a
+        - list of :term:`results dict`s, which represent fully-formed
+          experiments
 
         params: a dict of parameters for the experiment
-        returns: the experimental results """
+        returns: the experimental results
+
+        """
         return dict()
 
-    def report(self, params : Dict[str, Any], meta : Dict[str, Any], res : Union[Dict[str, Any], List[ResultsDict]]) -> ResultsDict:
-        """Return a properly-structured dict of results. The default returns a dict with
-        results keyed by :attr:`Experiment.RESULTS`, the data point in the parameter space
-        keyed by :attr:`Experiment.PARAMETERS`, and timing and other metadata keyed
-        by :attr:`Experiment.METADATA`. Overriding this method can be used to record extra
-        metadata values, but be sure to call the base method as well.
+    def report(self, params: Dict[str, Any], meta: Dict[str, Any], res: Union[Dict[str, Any], List[ResultsDict]]) -> ResultsDict:
+        """Return a properly-structured dict of results. The default returns a
+        dict with results keyed by :attr:`Experiment.RESULTS`, the
+        data point in the parameter space keyed by
+        :attr:`Experiment.PARAMETERS`, and timing and other metadata
+        keyed by :attr:`Experiment.METADATA`. Overriding this method
+        can be used to record extra metadata values, but be sure to
+        call the base method as well.
 
-        If the experimental results are a list of results dicts, then we report a
-        results dict whose results are a list of results dicts.
-        This is used by :class:`RepeatedExperiment` amd other experiments that want to report
-        multiple sets of results.
+        If the experimental results are a list of results dicts, then
+        we report a results dict whose results are a list of results
+        dicts.  This is used by :class:`RepeatedExperiment` amd other
+        experiments that want to report multiple sets of results.
 
         :param params: the parameters we ran under
         :param meta: the metadata for this run
         :param res: the direct experimental results from do()
-        :returns: a :term:`results dict`"""
+        :returns: a :term:`results dict`
+
+        """
         rc = Experiment.resultsdict()
         rc[self.PARAMETERS] = params.copy()
         rc[self.METADATA] = meta.copy()
         rc[self.RESULTS] = res
         return rc
 
-    def run(self) -> ResultsDict:
-        """Run the experiment, using the parameters set using :meth:`set`.
-        A "run" consists of calling :meth:`setUp`, :meth:`do`, and :meth:`tearDown`,
-        followed by collecting and storing (and returning) the
-        experiment's results. If running the experiment raises an
-        exception, that will be returned in the metadata along with
-        its traceback to help with experiment debugging.
+    def run(self, fatal: bool = False) -> ResultsDict:
+        """Run the experiment, using the parameters set using :meth:`set`.  A
+        "run" consists of calling :meth:`setUp`, :meth:`do`, and
+        :meth:`tearDown`, followed by collecting and storing (and
+        returning) the experiment's results. If running the experiment
+        raises an exception, that will be returned in the metadata
+        along with its traceback to help with experiment debugging.
+        If fatal is True it will also be raised from this method: the
+        default prints the exception but doesn't raise it.
+
+        :param fatal: (optional) raise any exceptions (default suppresses them into the results dict)
+        :returns: a :term:`results dict`
+
+        """
 
-        :returns: a :term:`results dict`"""
-        
         # record the experiment's class name
         self._metadata = dict()
-        cn = '{modulename}.{classname}'.format(modulename = self.__class__.__module__,
-                                               classname = self.__class__.__name__)
+        cn = '{modulename}.{classname}'.format(modulename=self.__class__.__module__,
+                                               classname=self.__class__.__name__)
         self._metadata[self.EXPERIMENT] = cn
 
         # perform the experiment protocol
@@ -228,22 +240,20 @@ class Experiment(object):
             elapsedTime += dt
             self._metadata[self.EXPERIMENT_TIME] = dt
             self.tearDown()
-            doneTeardownTime = datetime.now() 
+            doneTeardownTime = datetime.now()
             dt = (doneTeardownTime - doneExperimentTime).total_seconds()
             elapsedTime += dt
             self._metadata[self.TEARDOWN_TIME] = dt
             self._metadata[self.END_TIME] = doneTeardownTime
             self._metadata[self.ELAPSED_TIME] = elapsedTime
-            
+
             # set the success flag
             self._metadata[self.STATUS] = True
         except Exception as e:
-            print("Caught exception in experiment: {e}".format(e=e), file=sys.stderr)
-
             # grab the traceback before we do anything else
             #_, _, tb = sys.exc_info()
             tb = traceback.format_exc()
-            
+
             # decide on the cleanup actions that need doing
             if (doneSetupTime is not None) and (doneExperimentTime is None):
                 # we did the setup and then failed in the experiment, so
@@ -261,7 +271,13 @@ class Experiment(object):
             self._metadata[self.STATUS] = False
             self._metadata[self.EXCEPTION] = e
             self._metadata[self.TRACEBACK] = tb
-                
+
+            # propagate the exception or just report it
+            if fatal:
+                raise e
+            else:
+                print("Caught exception in experiment: {e}".format(e=e), file=sys.stderr)
+
         # report the results
         self._results = res
         return self.report(params,
@@ -271,34 +287,40 @@ class Experiment(object):
 
     # ---------- Accessing results ----------
 
-    def __getitem__(self, k : str) -> Any:
+    def __getitem__(self, k: str) -> Any:
         """Return the given element of the experimental results. This only
-        gives access to the experimental results, *not* to the parameters
-        or metadata.
+        gives access to the experimental results, *not* to the
+        parameters or metadata.
 
         :param k: the result key
         :returns: the value
-        :raises: KeyError if there is no such result"""
-        return (self.experimentalResults())[k]
-    
-    def success(self) -> bool:
-        """Test whether the experiment has been run successfully. This will
-        be False if the experiment hasn't been run, or if it's been run and
-        failed.
+        :raises: KeyError if there is no such result
 
-        :returns: ``True`` if the experiment has been run successfully"""
+        """
+        return (self.experimentalResults())[k]
+
+    def success(self) -> bool:
+        """Test whether the experiment has been run successfully. This will be
+        False if the experiment hasn't been run, or if it's been run
+        and failed.
+
+        :returns: ``True`` if the experiment has been run successfully
+
+        """
         if self.STATUS in self.metadata().keys():
             return (self.metadata())[self.STATUS]
         else:
             return False
 
     def failed(self) -> bool:
-        '''Test whether an experiment failed. This will be True if the experiment has
-        been run and has failed, which means that there will be an exception and traceback
-        information stored in the metadata. It will be False if the experiment hasn't
-        been run.
+        '''Test whether an experiment failed. This will be True if the
+        experiment has been run and has failed, which means that there
+        will be an exception and traceback information stored in the
+        metadata. It will be False if the experiment hasn't been run.
 
-        :returns: ``True`` if the experiment has failed'''
+        :returns: ``True`` if the experiment has failed
+
+        '''
         if self.STATUS not in self.metadata().keys():
             return False
         else:
@@ -314,28 +336,30 @@ class Experiment(object):
                            self.experimentalResults())
 
     def experimentalResults(self) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
-        """Return the experimental results from our last run. This will
-        be None if we haven't been run, or if we ran and failed.
+        """Return the experimental results from our last run. This will be
+        None if we haven't been run, or if we ran and failed.
 
-        :returns: the experimental results dict, which may be empty, and may be a list of dicts"""
+        :returns: the experimental results dict, which may be empty, and may be a list of dicts
+
+        """
         return self._results
-    
-    def parameters(self) -> Dict[str, Any]:
-        """Return the current experimental parameters, which will
-        be None if none have been given by a call to set()
 
-        :returns: the parameters, which may be empty"""
+    def parameters(self) -> Dict[str, Any]:
+        """Return the current experimental parameters, which will be None if
+        none have been given by a call to set()
+
+        :returns: the parameters, which may be empty
+
+        """
         return self._parameters
 
     def metadata(self) -> Dict[str, Any]:
-        """Return the metadata we collected at out last execution, which
-        will be None if we've not been executed and an empty dict if
-        we're mid-run (i.e., if this method is called from do() for
-        whatever reason).
+        """Return the metadata we collected at out last execution, which will
+        be None if we've not been executed and an empty dict if we're
+        mid-run (i.e., if this method is called from do() for whatever
+        reason).
 
-        :returns: the metadata, which may be empty"""
+        :returns: the metadata, which may be empty
+
+        """
         return self._metadata
-    
-    
-        
-    
