@@ -1,6 +1,6 @@
 # Simulation "lab notebook" for collecting results, in-memory version
 #
-# Copyright (C) 2016--2020 Simon Dobson
+# Copyright (C) 2016--2021 Simon Dobson
 #
 # This file is part of epyc, experiment management in Python.
 #
@@ -21,22 +21,21 @@ from epyc import Experiment, ResultSet, ResultsDict
 from pandas import DataFrame                               # type: ignore
 from contextlib import contextmanager
 import sys
+from typing import List, Set, Dict, Any, Optional, Union, cast
 if sys.version_info >= (3, 8):
-    from typing import List, Set, Dict, Any, Optional, Union, Final, cast
+    from typing import Final
 else:
-    # backwards compatibility with Python 35, Python36, and Python37 
-    from typing import List, Set, Dict, Any, Optional, Union, cast
     from typing_extensions import Final
 
 
 class ResultsStructureException(Exception):
     '''An exception raised when there is a problem with the structure of a
-    results dict. 
+    results dict.
 
-    :param rc: the results dict structure that causes the problem''' 
+    :param rc: the results dict structure that causes the problem'''
 
     def __init__(self, rc : Union[ResultsDict, List[ResultsDict]]):
-        super(ResultsStructureException, self).__init__("Can't handle results dict {rc}".format(rc=rc))
+        super().__init__("Can't handle results dict {rc}".format(rc=rc))
         self._rc = rc
 
     def resultsdict(self) -> Union[ResultsDict, List[ResultsDict]]:
@@ -54,16 +53,16 @@ class NotebookVersionException(Exception):
     :param actual: the actual version'''
 
     def __init__(self, expected : str, actual : str):
-        super(NotebookVersionException, self).__init__('Expected notebook version {v1}, got {v2}'.format(v1=expected, v2=actual))
+        super().__init__('Expected notebook version {v1}, got {v2}'.format(v1=expected, v2=actual))
         self._expected = expected
         self._actual = actual
-    
+
     def expectedVersion(self) -> str:
         '''Return the expected version of the file.
 
         :returns: the expected version'''
         return self._expected
-    
+
     def actualVersion(self) -> str:
         '''Return the actual version of the file.
 
@@ -76,33 +75,39 @@ class LabNotebookLockedException(Exception):
     attemoting to add result sets.'''
 
     def __init__(self):
-        super(LabNotebookLockedException, self).__init__('Lab notebook locked')
+        super().__init__('Lab notebook locked')
 
 
+class LabNotebook:
+    '''A "laboratory notebook" collecting together the results obtained
+    from different sets of experiments. A notebook is composed of
+    :class:`ResultSet` objects, which are homogeneous collections of
+    results of experiments performed at different values for the same
+    set of parameters. Each result set is tagged for access, with the
+    notebook using one result set as "current" at any time.
 
-class LabNotebook(object):
-    '''A "laboratory notebook" collecting together the results obtained from
-    different sets of experiments. A notebook is composed of :class:`ResultSet` objects,
-    which are homogeneous collections of results of experiments performed at different values
-    for the same set of parameters. Each result set is tagged for access,
-    with the notebook using one result set as "current" at any time.
+    The notebook collects together pending results from all result
+    sets so that they can be accessed uniformly. This is used by labs
+    to resolve pending results if there are multiple sets of
+    experiments running simultaneously.
 
-    The notebook collects together pending results from all result sets so that they
-    can be accessed uniformly. This is used by labs to resolve pending results if
-    there are multiple sets of experiments running simultaneously.
+    Result sets are immutable, but can be added and deleted freely
+    from notebooks: their contents cannot be changed, however.
 
     :param name: (optional) the notebook name (may be meaningful for sub-classes)
-    :param description: (optional) a free text description'''
- 
+    :param description: (optional) a free text description
+
+    '''
+
     # Defaults
     DEFAULT_RESULTSET : Final[str] = 'epyc.resultset.default'  #: Tag for the default result set.
 
     def __init__(self, name : str ='', description : str =None):
         if description is None:
-            description = 'A notebook' 
+            description = 'A notebook'
         self._name : Optional[str] = name                    # name
         self._description : str = description                # description
-        self._resultSets: Dict[str, ResultSet] = dict()      # tag to result set
+        self._resultSets : Dict[str, ResultSet] = dict()     # tag to result set
         self._resultSetTags : Dict[ResultSet, str] = dict()  # result set to tag
         self._pending : Dict[str, ResultSet] = dict()        # pending results job ids to result sets
         self._locked : bool = False                          # locked flag
@@ -173,6 +178,22 @@ class LabNotebook(object):
         self._current = rs
         return self._current
 
+    def deleteResultSet(self, rs : Union[str, ResultSet]):
+        '''Delete a result set.
+
+        :param rs: the result set or its tag'''
+        self.assertUnlocked()
+        if isinstance(rs, ResultSet):
+            rss = rs
+            tag = self._resultSetTags[rs]
+        else:
+            tag = rs
+            rss = self._resultSets[tag]
+        if rss == self._current:
+            raise Exception("Can't delete current result set ({tag})".format(tag=tag))
+        del self._resultSets[tag]
+        del self._resultSetTags[rss]
+
     def resultSets(self) -> List[str]:
         '''Return the tags for all the result sets in this notebook.
 
@@ -191,7 +212,7 @@ class LabNotebook(object):
 
         :param rs: the result set
         :returns: the tag'''
-        return self._resultSetTags[rs] 
+        return self._resultSetTags[rs]
 
     def select(self, tag: str) -> ResultSet:
         '''Select the given result set as current. Sub-classes may use this
@@ -199,9 +220,25 @@ class LabNotebook(object):
 
         :param tag: the tag
         :returns: the result set'''
-        self._current = self._resultSets[tag] 
+        self._current = self._resultSets[tag]
         return self._current
 
+    def already(self, tag : str, description : str =None) -> bool:
+        '''Check whether a result set exists. If it does, select it
+        and return True; if it doesn't, add it and return False.
+        This is a single-call combination of :meth:`contains` and
+        :meth:`select` that's useful for avoiding repeated computation.
+
+        :param tag: the result set tag
+        :param description: (optional) description if a result set is created
+        :returns: True if the set existed'''
+        if tag in self:
+            self.select(tag)
+            return True
+        else:
+            self.addResultSet(tag, description=description)
+            return False
+
     def current(self) -> ResultSet:
         '''Return the current result set.
 
@@ -227,18 +264,32 @@ class LabNotebook(object):
         :returns: the number of result sets'''
         return self.numberOfResultSets()
 
+    def keys(self) -> List[str]:
+        '''Return the result set tags in this notebook. The same
+        as :meth:`resultSets`.
+
+        :returns: the result set tags'''
+        return self.resultSets()
+
+    def __contains__(self, tag : str) -> bool:
+        '''Tests if the given result set ic contained in this notebook.
+
+        :param tag: the result set tag
+        :returns: True if the result set exists'''
+        return (tag in self.resultSets())
+
 
     # ---------- Finishing ----------
 
     def finish(self, commit=True):
         '''Mark the entire notebook as finished, closing and locking all result sets
         against further changes. Finishing a persistent notebook commits it.
-        
+
         By default the finished notebook is committed as such. In certain cases
         it may be desirable to finish the notebook but not commit it, *i.e.*, to
         stop updates in memory without changing the backing file. Setting ``commit=False``
         will accomplish this.
-        
+
         :param commit: (optional) commit the notebook (defaults to True)'''
         if not self.isLocked():
             for tag in self.resultSets():
@@ -309,7 +360,7 @@ class LabNotebook(object):
         rs.resolveSinglePendingResult(jobid)
 
         # mark the job as resolved with the notebook
-        del self._pending[jobid] 
+        del self._pending[jobid]
 
     def cancelPendingResult(self, jobid : str):
         '''Cancel the given pending result.
@@ -328,10 +379,10 @@ class LabNotebook(object):
         rs.cancelSinglePendingResult(jobid)
 
         # mark the job as resolved with the notebook
-        del self._pending[jobid] 
-            
+        del self._pending[jobid]
+
     def ready(self, tag : str =None) -> bool:
-        '''Test whether the result set has pending results. 
+        '''Test whether the result set has pending results.
 
         :params tag: (optional) the result set tag (defaults to the current set)
         :returns: True if all pending results have been resolved (or cancelled)'''
@@ -343,7 +394,7 @@ class LabNotebook(object):
 
     def readyFraction(self, tag : str =None) -> float:
         """Test what fraction of results are available in the tagged result set.
-        
+
         :params tag: (optional) the result set tag (defaults to the current set)
         :returns: the fraction of available results"""
         if tag is None:
@@ -361,7 +412,7 @@ class LabNotebook(object):
 
     def pendingResults(self, tag : str =None) -> List[str]:
         '''Return the identifiers of the results pending in the tagged dataset.
-        
+
         :params tag: (optional) the result set tag (defaults to the current set)
         :returns: a set of job identifiers'''
         if tag is None:
@@ -372,7 +423,7 @@ class LabNotebook(object):
 
     def numberOfPendingResults(self, tag : str =None) -> int:
         '''Return the number of results pending in the tagged dataset.
-        
+
         :params tag: (optional) the result set tag (defaults to the current set)
         :returns: the number of results'''
         if tag is None:
@@ -402,7 +453,7 @@ class LabNotebook(object):
         '''Return the number of results pending in all result sets.
 
         :returns: the total number of pending results'''
-        return len(self._pending) 
+        return len(self._pending)
 
 
     # --------- Managing results ----------
@@ -432,7 +483,7 @@ class LabNotebook(object):
         One may also add a list of results dicts, in which case they will
         be added individually.
 
-        Any structure of results dicts that can't be handled will raise a 
+        Any structure of results dicts that can't be handled will raise a
         :class:`ResultsStructureException`.
 
         :param result: a results dict or collection of them
@@ -456,17 +507,17 @@ class LabNotebook(object):
 
         else:
             raise ResultsStructureException(results)
-    
+
     def dataframe(self, tag : str =None, only_successful : bool =True) -> DataFrame:
         """Return results as a ``pandas.DataFrame``. If no tag is provided,
-        use the current result set. 
+        use the current result set.
 
         If the only_successful flag is set (the default), then the DataFrame will
         only include results that completed without an exception; if it is set to
         False, the DataFrame will include all results and also the exception details.
 
         If you are only interested in results corresponding to some sets of parameters
-        you can pre-filter the dataframe using :meth:`dataframeFor`. 
+        you can pre-filter the dataframe using :meth:`dataframeFor`.
 
         :params tag: (optional) the tag of the result set (defaults to the currently select result set)
         :param only_successful: include only successful experiments (defaults to True)
@@ -480,7 +531,7 @@ class LabNotebook(object):
 
     def dataframeFor(self, params : Dict[str, Any], tag : str =None, only_successful : bool =True) -> DataFrame:
         """Return results for the goven parameter values as a ``pandas.DataFrame``.
-        If no tag is provided, the current result set is queried. 
+        If no tag is provided, the current result set is queried.
         If the only_successful flag is set (the default), then the DataFrame will
         only include results that completed without an exception; if it is set to
         False, the DataFrame will include all results and also the exception details.
@@ -530,7 +581,7 @@ class LabNotebook(object):
 
     def numberOfResults(self, tag : str =None) -> int:
         '''Return the number of results in the tagged dataset.
-        
+
         :params tag: (optional) the result set tag (defaults to the current set)
         :returns: the number of results'''
         if tag is None:
@@ -553,5 +604,3 @@ class LabNotebook(object):
         finally:
             # commit any changes and close the file
             self.commit()
-
-
