@@ -326,9 +326,71 @@ class LabNotebookTests(unittest.TestCase):
         self.assertEqual(len(rcs), 1)
         self.assertFalse(rcs[0][Experiment.METADATA][Experiment.STATUS])
 
+    def testDeleteResultSetByTag(self):
+        '''Test we can delete a result set.'''
+        rs = self._nb.addResultSet('second')
+        self._nb.deleteResultSet(LabNotebook.DEFAULT_RESULTSET)
+        self.assertEqual(len(self._nb), 1)
+        self.assertEqual(self._nb.currentTag(), 'second')
+        self.assertCountEqual(self._nb.resultSets(), ['second'])
+        self.assertEqual(self._nb.current(), rs)
+
+    def testDeleteResultSetByReference(self):
+        '''Test we can delete a result set by reference.'''
+        rs = self._nb.addResultSet('second')
+        self._nb.addResultSet('third')
+        self._nb.deleteResultSet(rs)
+        self.assertEqual(len(self._nb), 2)
+        self.assertEqual(self._nb.currentTag(), 'third')
+        self.assertCountEqual(self._nb.resultSets(), [LabNotebook.DEFAULT_RESULTSET, 'third'])
+
+    def testDeleteCurrent(self):
+        '''Test we can't delete the current result set.''' 
+        rs = self._nb.addResultSet('second')
+        with self.assertRaises(Exception):
+            self._nb.deleteResultSet(rs)
+
+    def testKeys(self):
+        '''Test we can extract the result set tags.'''
+        self.assertCountEqual(self._nb.keys(), [LabNotebook.DEFAULT_RESULTSET])
+        self._nb.addResultSet('second')
+        self._nb.addResultSet('third')
+        self.assertCountEqual(self._nb.keys(), [LabNotebook.DEFAULT_RESULTSET, 'second', 'third'])
+
+    def testContains(self):
+        '''Test we can check containment.'''
+        self._nb.addResultSet('second')
+        self.assertTrue('second' in self._nb)
+        self.assertTrue(LabNotebook.DEFAULT_RESULTSET in self._nb)
+        self.assertFalse('third' in self._nb)
+
+        self._nb.select(LabNotebook.DEFAULT_RESULTSET)
+        self._nb.deleteResultSet('second')
+        self.assertFalse('second' in self._nb)
+        self.assertTrue(LabNotebook.DEFAULT_RESULTSET in self._nb)
+
+    def testAlready(self):
+        '''Test we can test and select in one method call.'''
+        self.assertEqual(self._nb.currentTag(), LabNotebook.DEFAULT_RESULTSET)
+
+        # first call adds and selects the result set 
+        self.assertFalse(self._nb.already('second'))
+        self.assertEqual(self._nb.currentTag(), 'second')
+
+        # add a result to make sure it's maintained
+        rc1 = self._resultsdict()
+        rc1[Experiment.PARAMETERS]['a'] = 10
+        self._nb.addResult(rc1)
+
+        # second call just selects it
+        self.assertTrue(self._nb.already('second'))
+        self.assertEqual(self._nb.currentTag(), 'second')
+        self.assertEqual(len(self._nb.current()), 1)
+
+        
 # TODO: Test we can add metadata
 
 if __name__ == '__main__':
     unittest.main()
 
- 
\ No newline at end of file
+ 
